#include <algorithm>
#include <msp/strings/format.h>
#include <IL/il.h>
#include "devilloader.h"

using namespace std;

namespace {

unsigned char eof(void *handle)
{
	return reinterpret_cast<Msp::IO::Seekable *>(handle)->eof();
}

int get(void *handle)
{
	return reinterpret_cast<Msp::IO::Seekable *>(handle)->get();
}

int read(void *buf, unsigned size, unsigned count, void *handle)
{
	Msp::IO::Seekable *io = reinterpret_cast<Msp::IO::Seekable *>(handle);
	char *cbuf = reinterpret_cast<char *>(buf);
	unsigned i = 0;
	for(; i<count; ++i)
	{
		unsigned len = io->read(cbuf, size);
		cbuf += size;
		if(len<size)
			break;
	}
	return i;
}

int seek(void *handle, int offset, int il_type)
{
	Msp::IO::SeekType type;
	if(il_type==IL_SEEK_SET)
		type = Msp::IO::S_BEG;
	else if(il_type==IL_SEEK_CUR)
		type = Msp::IO::S_CUR;
	else if(il_type==IL_SEEK_END)
		type = Msp::IO::S_END;
	else
		return -1;

	reinterpret_cast<Msp::IO::Seekable *>(handle)->seek(offset, type);

	return 0;
}

int tell(void *handle)
{
	return reinterpret_cast<Msp::IO::Seekable *>(handle)->tell();
}

std::string error_string(ILenum err)
{
	switch(err)
	{
	case IL_FORMAT_NOT_SUPPORTED: return "Format not supported";
	case IL_INTERNAL_ERROR: return "DevIL internal error";
	case IL_INVALID_FILE_HEADER: return "Invalid file header";
	case IL_FILE_READ_ERROR: return "File read error";
	case IL_LIB_PNG_ERROR: return "LibPNG error";
	case IL_LIB_JPEG_ERROR: return "LibJPEG error";
	default: return Msp::format("Unknown error (%04X)", err);
	}
}

} // namespace


namespace Msp {
namespace Graphics {

DevilLoader::DevilLoader(IO::Seekable &i):
	io(i)
{
	static bool il_init_done = false;

	if(!il_init_done)
	{
		ilInit();
		ilEnable(IL_ORIGIN_SET);
		ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
		il_init_done = true;
	}

	ilGenImages(1, &id);
}

DevilLoader::~DevilLoader()
{
	ilDeleteImages(1, &id);
}

bool DevilLoader::detect(const string &sig)
{
	unsigned type = ilDetermineTypeL(sig.data(), sig.size());
	return type!=IL_TYPE_UNKNOWN;
}

void DevilLoader::load(Image::Data &data)
{
	ilSetRead(0, 0, eof, get, read, seek, tell);
	ilBindImage(id);

	ilGetError();
	if(!ilLoadF(IL_TYPE_UNKNOWN, &io))
		throw bad_image_data(error_string(ilGetError()));

	switch(ilGetInteger(IL_IMAGE_FORMAT))
	{
	case IL_COLOR_INDEX:     data.fmt = COLOR_INDEX; break;
	case IL_LUMINANCE:       data.fmt = LUMINANCE; break;
	case IL_LUMINANCE_ALPHA: data.fmt = LUMINANCE_ALPHA; break;
	case IL_RGB:             data.fmt = RGB; break;
	case IL_RGBA:            data.fmt = RGBA; break;
	case IL_BGR:             data.fmt = BGR; break;
	case IL_BGRA:            data.fmt = BGRA; break;
	default: throw unsupported_image_format("unknown pixel format");
	}

	data.width = ilGetInteger(IL_IMAGE_WIDTH);
	data.height = ilGetInteger(IL_IMAGE_HEIGHT);
	data.stride = data.width*ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL);
	unsigned data_size = data.stride*data.height;
	data.pixels = new char[data_size];
	ILubyte *il_data = ilGetData();
	copy(il_data, il_data+data_size, data.pixels);

	ilBindImage(0);
	ilResetRead();
}

} // namespace Graphics
} // namespace Msp
